/*
 * Client shell daemon v1.2
 *
 * (c) Copyright 1998-1999 Stelian Pop <pop@cybercable.fr>.
 *
 * This is distributed under the GNU GENERAL PUBLIC LICENSE (GPL) Version 2
 *
*/
#include <sys/socket.h>
#include <netinet/in.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <signal.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <pwd.h>
#include <netdb.h>

struct passwd *pwd;
struct sockaddr_in addr;
struct hostent *hp;
int sk_serv, sk_cli, port, lg;

int getstr(int desc, char *buf, int cnt) {
  char c;
  
  do {
    if (read(desc, &c, 1) != 1)
      return -1;
    *buf++ = c;
    if (--cnt == 0) {
      return -1;
    }
  } while (c != 0);
  return 0;
}

int main(int argc, char *argv[]) {
  int pid;
  char buffer[1024];
  char luser[1024];

  /* Create the server socket */
  sk_serv=socket(AF_INET, SOCK_STREAM, 0);
  if (sk_serv==-1) {
    perror("socket error");
    exit(1);
  }

  /* Listen on the socket */
  if (listen(sk_serv,20)==-1) {
    perror("listen error");
    exit(1);
  }

  lg = sizeof(addr);
  if (getsockname(sk_serv, (struct sockaddr *)&addr, &lg)==-1) {
    perror("getsockname error");
    exit(1);
  }

  fprintf(stdout, "OK: %d\n", ntohs(addr.sin_port));
  fflush(stdout);

  close(0);
  close(1);
  close(2);
  
  setsid();
  setpgrp();
  pid = fork();
  if (pid == -1) {
    perror("fork error");
    exit(1);
  }
  if (pid != 0)
    exit(0);

 
  boucle:

  /* Connect loop */
  while (1) {
    
    /* Accept the connection */
    lg=sizeof(addr);
    do 
      sk_cli=accept(sk_serv, (struct sockaddr *)&addr, &lg);
    while ((sk_cli==-1)&&(errno==EINTR));
    if (sk_cli==-1) {
      close(sk_cli);
      sk_cli = 0;
      goto boucle;
    }

    /* Read the local user */
    pwd = getpwuid(getuid());
    if (pwd == NULL) {
      strcpy(buffer, "User unknown\n");
      write(sk_cli, buffer, sizeof(buffer));
      close(sk_cli);
      goto boucle;
    }
    strncpy(luser, pwd->pw_name,1024);

    /* Read the remote user */
    if (getstr(sk_cli, buffer, sizeof(buffer)) == -1) {
      strcpy(buffer, "Read error\n");
      write(sk_cli, buffer, sizeof(buffer));
      close(sk_cli);
      goto boucle;
    }

    /* Read the remote machine name */
    hp = gethostbyaddr((const char *)&addr.sin_addr,
                       sizeof(struct in_addr),
                       AF_INET);
    if (!hp || !hp->h_name) {
      strcpy(buffer, "Unable to find the remote machine\n");
      write(sk_cli, buffer, sizeof(buffer));
      close(sk_cli);
      goto boucle;
    }

    /* Test the rights */
    if (ruserok(hp->h_name, 0, buffer, luser)==-1) {
      strcpy(buffer, "Acces denied\n");
      write(sk_cli, buffer, sizeof(buffer));
      close(sk_cli);
      goto boucle;
    }

    /* Read the command line */
    if (getstr(sk_cli, buffer, sizeof(buffer)) == -1) {
      strcpy(buffer, "Read error\n");
      write(sk_cli, buffer, sizeof(buffer));
      close(sk_cli);
      goto boucle;
    }

    /* Stop the server? */
    if (!strcmp(buffer, "Arret ")) {
      strcpy(buffer, "Server stopped, bye.\n");
      write(sk_cli, buffer, sizeof(buffer));
      close(sk_cli);
      close(sk_serv);
      exit(0);
    }

    /* Redirection of descriptors */
    dup2(sk_cli, 0);
    dup2(sk_cli, 1);
    dup2(sk_cli, 2);
    
    /* Execution of the command */
    if (system(buffer) == -1) {
      strcpy(buffer, "Exec error\n");
      write(sk_cli, buffer, sizeof(buffer));
      close(sk_cli);
      close(0);
      close(1);
      close(2);
      goto boucle;
    }

    /* Close the connection */
    close(sk_cli);
    close(0);
    close(1);
    close(2);
    sk_cli = 0;
  }
}
